@@ -1,16 +1,13 @@
 # -*- coding: utf-8 -*-
 from django.core.urlresolvers import reverse
-from django.db.models.fields import FieldDoesNotExist
-from django.template import RequestContext, Context, Template
 from django.utils.encoding import force_unicode, StrAndUnicode
 from django.utils.datastructures import SortedDict
 from django.utils.text import capfirst
-from django.utils.html import escape
-from django.utils.safestring import mark_safe, SafeData
-from itertools import ifilter, islice
-import warnings
-from .templatetags.django_tables2 import title
-from .utils import A, AttributeDict, Attrs, OrderBy, Sequence
+from django.utils.safestring import mark_safe
+from django.template import RequestContext, Context, Template
+from django.db.models.fields import FieldDoesNotExist
+from .utils import OrderBy, A, AttributeDict
+from itertools import ifilter
 
 
 class Column(object):
@@ -20,50 +17,41 @@ class Column(object):
     :class:`Column` objects control the way a column (including the cells that
     fall within it) are rendered.
 
-    :param verbose_name: A human readable version of the column name. This
-                         should not be title case. It is converted to title
-                         case for use in column headers.
-    :type  verbose_name: ``unicode``
-    :type      accessor: :class:`basestring` or :class:`~.utils.Accessor`
-    :param     accessor: An accessor that describes how to extract values for
-                         this column from the :term:`table data`.
-    :parameter  default: The default value for the column. This can be a value
-                         or a callable object [1]_. If an object in the data
-                         provides :const:`None` for a column, the default will
-                         be used instead.
+    :param verbose_name: A pretty human readable version of the column name.
+        Typically this is used in the header cells in the HTML output.
 
-                         The default value may affect ordering, depending on
-                         the type of data the table is using. The only case
-                         where ordering is not affected is when a
-                         :class:`QuerySet` is used as the table data (since
-                         sorting is performed by the database).
+    :type accessor: :class:`basestring` or :class:`~.utils.Accessor`
+    :param accessor: An accessor that describes how to extract values for this
+        column from the :term:`table data`.
 
-                         .. [1] The provided callable object must not expect to
-                                receive any arguments.
-    :type      visible: :class:`bool`
-    :param     visible: If :const:`False`, this column will not be in HTML from
-                        output generators (e.g. :meth:`as_html` or
-                        ``{% render_table %}``).
+    :param default: The default value for the column. This can be a value or a
+        callable object [1]_. If an object in the data provides :const:`None`
+        for a column, the default will be used instead.
 
-                        When a field is not visible, it is removed from the
-                        table's :attr:`~Column.columns` iterable.
-    :type     sortable: :class:`bool`
-    :param    sortable: If :const:`False`, this column will not be allowed to
-                        influence row ordering/sorting.
-    :type        attrs: :class:`Attrs` object
-    :param       attrs: HTML attributes to be added to components in the column
+        The default value may affect ordering, depending on the type of
+        data the table is using. The only case where ordering is not
+        affected ing when a :class:`QuerySet` is used as the table data
+        (since sorting is performed by the database).
 
-    Supported ``Attrs`` keys are:
+        .. [1] The provided callable object must not expect to receive any
+           arguments.
 
-    - *th* -- ``<th>`` element in header
-    - *td* -- ``<td>`` element in body
-    - *cell* -- fall back for ``<th>`` and ``<td>`` should they not be specified
+    :type visible: :class:`bool`
+    :param visible: If :const:`False`, this column will not be in HTML from
+        output generators (e.g. :meth:`as_html` or ``{% render_table %}``).
+
+        When a field is not visible, it is removed from the table's
+        :attr:`~Column.columns` iterable.
+
+    :type sortable: :class:`bool`
+    :param sortable: If :const:`False`, this column will not be allowed to
+        influence row ordering/sorting.
     """
     #: Tracks each time a Column instance is created. Used to retain order.
     creation_counter = 0
 
     def __init__(self, verbose_name=None, accessor=None, default=None,
-                 visible=True, sortable=None, attrs=None):
+                 visible=True, sortable=None):
         if not (accessor is None or isinstance(accessor, basestring) or
                 callable(accessor)):
             raise TypeError(u'accessor must be a string or callable, not %s' %
@@ -76,12 +64,6 @@ class Column(object):
         self.sortable = sortable
         self.verbose_name = verbose_name
         self.visible = visible
-        attrs = attrs or {}
-        if not isinstance(attrs, Attrs):
-            warnings.warn('attrs must be Attrs object, not %s'
-                          % attrs.__class__.__name__, DeprecationWarning)
-            attrs = Attrs(attrs)
-        self.attrs = attrs
 
         self.creation_counter = Column.creation_counter
         Column.creation_counter += 1
@@ -101,10 +83,7 @@ class Column(object):
         """
         The value used for the column heading (e.g. inside the ``<th>`` tag).
 
-        By default this titlises the column's :attr:`verbose_name`. If
-        ``verbose_name`` is an instance of ``SafeData``, it's used unmodified.
-
-        :returns: ``unicode`` or ``None``
+        By default this equivalent to the column's :attr:`verbose_name`.
 
         .. note::
 
@@ -115,14 +94,9 @@ class Column(object):
             object hence accessing that first) when this property doesn't
             return something useful.
         """
-        if self.verbose_name:
-            if isinstance(self.verbose_name, SafeData):
-                # If the author has used mark_safe, we're going to assume the
-                # author wants the value used verbatim.
-                return self.verbose_name
-            return title(self.verbose_name)
+        return self.verbose_name
 
-    def render(self, value):
+    def render(self, value, **kwargs):
         """
         Returns the content for a specific cell.
 
@@ -146,61 +120,45 @@ class CheckBoxColumn(Column):
 
     This class implements some sensible defaults:
 
-    - The ``name`` attribute of the input is the :term:`column name` (can
-      override via ``attrs`` argument).
+    - The ``name`` attribute of the input is the name of the :term:`column
+      name` (can be overriden via ``attrs`` argument).
     - The ``sortable`` parameter defaults to :const:`False`.
+    - The ``type`` attribute of the input is ``checkbox`` (can be overriden via
+      ``attrs`` argument).
+    - The header checkbox is left bare, i.e. ``<input type="checkbox"/>`` (use
+      the ``header_attrs`` argument to customise).
 
-    .. note::
+    .. note:: The "apply some operation onto the selection" functionality is
+        not implemented in this column, and requires manually implemention.
 
-        You'd expect that you could select multiple checkboxes in the rendered
-        table and then *do something* with that. This functionality isn't
-        implemented. If you want something to actually happen, you'll need to
-        implement that yourself.
-
-    In addition to ``Attrs`` keys supported by ``Column``, the following are
-    available:
-
-    - *input*     -- ``<input>`` elements in both ``<td>`` and ``<th>``.
-    - *th__input* -- If defined: used *instead of* ``input`` in table header.
-    - *td__input* -- If defined: used *instead of* ``input`` in table body.
+    :param attrs:
+        a :class:`dict` of HTML attributes that are added to the rendered
+        ``<input type="checkbox" .../>`` tag
+    :param header_attrs:
+        same as *attrs*, but applied **only** to the header checkbox
     """
-    def __init__(self, attrs=None, **extra):
-        header_attrs = extra.pop('header_attrs', None)
-        # For backwards compatibility, passing in a normal dict effectively
-        # should assign attributes to the `<input>` tag.
-        attrs = attrs or Attrs()
-        if not isinstance(attrs, Attrs):
-            warnings.warn('attrs must be Attrs object, not %s'
-                          % attrs.__class__.__name__, DeprecationWarning)
-            attrs = Attrs(td__input=attrs)
-        # This is done for backwards compatible too, there used to be a
-        # ``header_attrs`` argument, but this has been deprecated. We'll
-        # maintain it for a while by translating it into ``head.checkbox``.
-        if header_attrs:
-            attrs.setdefault('th__input', header_attrs)
-
-        kwargs = {'sortable': False, 'attrs': attrs}
-        kwargs.update(extra)
-        super(CheckBoxColumn, self).__init__(**kwargs)
-
+    def __init__(self, attrs=None, header_attrs=None, **extra):
+        params = {'sortable': False}
+        params.update(extra)
+        super(CheckBoxColumn, self).__init__(**params)
+        self.attrs = attrs or {}
+        self.header_attrs = header_attrs or {}
 
     @property
     def header(self):
-        default = {'type': 'checkbox'}
-        general = self.attrs.get('input')
-        specific = self.attrs.get('th__input')
-        attrs = AttributeDict(default, **(specific or general or {}))
+        attrs = AttributeDict({
+            'type': 'checkbox',
+        })
+        attrs.update(self.header_attrs)
         return mark_safe(u'<input %s/>' % attrs.as_html())
 
     def render(self, value, bound_column):
-        default = {
+        attrs = AttributeDict({
             'type': 'checkbox',
             'name': bound_column.name,
             'value': value
-        }
-        general = self.attrs.get('input')
-        specific = self.attrs.get('td__input')
-        attrs = AttributeDict(default, **(specific or general or {}))
+        })
+        attrs.update(self.attrs)
         return mark_safe(u'<input %s/>' % attrs.as_html())
 
 
@@ -216,13 +174,15 @@ class LinkColumn(Column):
     described. The last argument ``attrs`` allows custom HTML attributes to
     be added to the ``<a>`` tag.
 
-    :param    viewname: See :func:`django.core.urlresolvers.reverse`.
-    :param     urlconf: See :func:`django.core.urlresolvers.reverse`.
-    :param        args: See :func:`django.core.urlresolvers.reverse`. **
-    :param      kwargs: See :func:`django.core.urlresolvers.reverse`. **
+    :param viewname: See :func:`django.core.urlresolvers.reverse`.
+    :param urlconf: See :func:`django.core.urlresolvers.reverse`.
+    :param args: See :func:`django.core.urlresolvers.reverse`. **
+    :param kwargs: See :func:`django.core.urlresolvers.reverse`. **
     :param current_app: See :func:`django.core.urlresolvers.reverse`.
-    :param       attrs: a :class:`dict` of HTML attributes that are added to
-                        the rendered ``<input type="checkbox" .../>`` tag
+
+    :param attrs:
+        a :class:`dict` of HTML attributes that are added to the rendered
+        ``<input type="checkbox" .../>`` tag
 
     ** In order to create a link to a URL that relies on information in the
     current row, :class:`.Accessor` objects can be used in the ``args`` or
@@ -247,38 +207,19 @@ class LinkColumn(Column):
 
         class PeopleTable(tables.Table):
             name = tables.LinkColumn('people_detail', args=[A('pk')])
-
-    In addition to ``Attrs`` keys supported by ``Column``, the following are
-    available:
-
-    - *a* -- ``<a>`` elements in ``<td>``.
     """
     def __init__(self, viewname, urlconf=None, args=None, kwargs=None,
                  current_app=None, attrs=None, **extra):
-        # backwards compatible translation for naive attrs value
-        attrs = attrs or Attrs()
-        if not isinstance(attrs, Attrs):
-            warnings.warn('attrs must be Attrs object, not %s'
-                          % attrs.__class__.__name__, DeprecationWarning)
-            attrs = Attrs(a=attrs)
-        extra['attrs'] = attrs
         super(LinkColumn, self).__init__(**extra)
         self.viewname = viewname
         self.urlconf = urlconf
         self.args = args
         self.kwargs = kwargs
         self.current_app = current_app
+        self.attrs = attrs or {}
 
     def render(self, value, record, bound_column):
-        # Remember that value is actually what would have normally been put
-        # into the cell. i.e. it *already* takes into consideration the
-        # column's *default* property, thus we must check the actual data value
-        # and use that to decide whether to render a link or just the default
-        try:
-            raw = bound_column.accessor.resolve(record)
-        except (TypeError, AttributeError, KeyError, ValueError) as e:
-            raw = None
-        if raw is None:
+        if value is None:
             return self.default
         # The following params + if statements create the arguments required to
         # pass to Django's reverse() function.
@@ -295,20 +236,20 @@ class LinkColumn(Column):
             params['args'] = [a.resolve(record) if isinstance(a, A) else a
                               for a in self.args]
         if self.kwargs:
-            params['kwargs'] = {}
-            for k, v in self.kwargs.items():
-                # If we're dealing with an Accessor (A), resolve it, otherwise
-                # use the value verbatim.
-                params['kwargs'][k] = (v.resolve(record) if isinstance(v, A)
-                                       else v)
+            params['kwargs'] = self.kwargs
+            for key, value in self.kwargs:
+                if isinstance(value, A):
+                    params['kwargs'][key] = value.resolve(record)
         if self.current_app:
-            params['current_app'] = (self.current_app.resolve(record)
-                                     if isinstance(self.current_app, A)
-                                     else self.current_app)
+            params['current_app'] = self.current_app
+            for key, value in self.current_app:
+                if isinstance(value, A):
+                    params['current_app'][key] = value.resolve(record)
+        url = reverse(**params)
         html = u'<a href="{url}" {attrs}>{value}</a>'.format(
-            url=escape(reverse(**params)),
-            attrs=AttributeDict(self.attrs.get('a', {})).as_html(),
-            value=escape(value)
+            url=reverse(**params),
+            attrs=AttributeDict(self.attrs).as_html(),
+            value=value
         )
         return mark_safe(html)
 
@@ -346,41 +287,40 @@ class TemplateColumn(Column):
         self.template_code = template_code
 
     def render(self, record, table, **kwargs):
-        # If the table is being rendered using `render_table`, it hackily
-        # attaches the context to the table as a gift to `TemplateColumn`. If
-        # the table is being rendered via `Table.as_html`, this won't exist.
-        context = getattr(table, 'context', Context())
-        context.update({'record': record})
-        try:
-            return Template(self.template_code).render(context)
-        finally:
-            context.pop()
+        t = Template(self.template_code)
+        if hasattr(table, 'request'):
+            context = RequestContext(table.request, {'record': record})
+        else:
+            context = Context({'record': record})
+        return t.render(context)
 
 
 class BoundColumn(object):
     """
-    A *run-time* version of :class:`.Column`. The difference between
-    ``BoundColumn`` and ``Column``, is that ``BoundColumn`` objects include the
-    relationship between a ``Column`` and a :class:`.Table`. In practice, this
-    means that a ``BoundColumn`` knows the *"variable name"* given to the
-    ``Column`` when it was declared on the ``Table``.
+    A *runtime* version of :class:`.Column`. The difference between
+    ``BoundColumn`` and ``Column``, is that ``BoundColumn`` objects are of the
+    relationship between a ``Column`` and a :class:`.Table`. This means that it
+    knows the *name* given to the ``Column``.
 
     For convenience, all :class:`.Column` properties are available from this
     class.
 
-    :type   table: :class:`.Table` object
-    :param  table: the table in which this column exists
-    :type  column: :class:`.Column` object
+    :type table: :class:`.Table` object
+    :param table: the table in which this column exists
+
+    :type column: :class:`.Column` object
     :param column: the type of column
-    :type    name: ``basestring`` object
-    :param   name: the variable name of the column used to when defining the
-                   :class:`.Table`. In this example the name is ``age``:
 
-                       .. code-block:: python
+    :type name: ``basestring`` object
+    :param name: the variable name of the column used to when defining the
+        :class:`.Table`. Example:
 
-                           class SimpleTable(tables.Table):
-                               age = tables.Column()
+        .. code-block:: python
 
+            class SimpleTable(tables.Table):
+                age = tables.Column()
+
+        ``age`` is the name.
     """
     def __init__(self, table, column, name):
         self._table = table
@@ -388,7 +328,7 @@ class BoundColumn(object):
         self._name = name
 
     def __unicode__(self):
-        return unicode(self.header)
+        return self.verbose_name
 
     @property
     def accessor(self):
@@ -398,38 +338,6 @@ class BoundColumn(object):
         """
         return self.column.accessor or A(self.name)
 
-    @property
-    def attrs(self):
-        """
-        Proxy to ``Column.attrs`` but injects some values of our own.
-
-        A ``th`` and ``td`` are guaranteed to be defined (irrespective of
-        what's actually defined in the column attrs. This makes writing
-        templates easier.
-        """
-        # Work on a copy of the Attrs object since we're tweaking stuff
-        attrs = dict(self.column.attrs)
-
-        # Find the relevant th attributes (fall back to cell if th isn't
-        # explicitly specified).
-        attrs["td"] = td = AttributeDict(attrs.get('td', attrs.get('cell', {})))
-        attrs["th"] = th = AttributeDict(attrs.get("th", attrs.get("cell", {})))
-        # make set of existing classes.
-        th_class = set((c for c in th.get("class", "").split(" ") if c))
-        td_class = set((c for c in td.get("class", "").split(" ") if c))
-        # add classes for ordering
-        if self.sortable:
-            th_class.add("sortable")
-        order_by = self.order_by
-        if order_by:
-            th_class.add("desc" if order_by.is_descending else "asc")
-        # Always add the column name as a class
-        th_class.add(self.name)
-        td_class.add(self.name)
-        if th_class: th['class'] = " ".join(sorted(th_class))
-        if td_class: td['class'] = " ".join(sorted(td_class))
-        return attrs
-
     @property
     def column(self):
         """
@@ -449,18 +357,7 @@ class BoundColumn(object):
         """
         The value that should be used in the header cell for this column.
         """
-        # favour Column.header
-        column_header = self.column.header
-        if column_header:
-            return column_header
-        # fall back to automatic best guess
-        verbose_name = self.verbose_name  # avoid calculating multiple times
-        if isinstance(verbose_name, SafeData):
-            # If the verbose_name has come from a model field, it's possible
-            # that the author used mark_safe to include HTML in the value. If
-            # this is the case, we leave it verbatim.
-            return verbose_name
-        return title(verbose_name)
+        return self.column.header or self.verbose_name
 
     @property
     def name(self):
@@ -476,8 +373,8 @@ class BoundColumn(object):
         instance, otherwise ``None``.
         """
         try:
-            return (self.table.order_by or {})[self.name]
-        except KeyError:
+            return self.table.order_by[self.name]
+        except IndexError:
             return None
 
     @property
@@ -507,19 +404,15 @@ class BoundColumn(object):
         then get the last field in the accessor (i.e. stop when the
         relationship turns from ORM relationships to object attributes [e.g.
         person.upper should stop at person]).
-
-        If the model field's ``verbose_name`` is a ``SafeData``, it's used
-        unmodified.
         """
-        # Favor an explicit defined verbose_name
+        # Favor an explicit verbose_name
         if self.column.verbose_name:
             return self.column.verbose_name
 
-        # This is our reasonable fallback, should the next section not result
-        # in anything useful.
+        # Reasonable fallback
         name = self.name.replace('_', ' ')
 
-        # Try to use a tmodel field's verbose_name
+        # Perhap use a model field's verbose_name
         if hasattr(self.table.data, 'queryset'):
             model = self.table.data.queryset.model
             parts = self.accessor.split('.')
@@ -535,7 +428,7 @@ class BoundColumn(object):
                 break
             if field:
                 name = field.verbose_name
-        return name
+        return capfirst(name)
 
     @property
     def visible(self):
@@ -563,11 +456,29 @@ class BoundColumns(object):
     At the moment you'll only come across this class when you access a
     :attr:`.Table.columns` property.
 
-    :type  table: :class:`.Table` object
+    :type table: :class:`.Table` object
     :param table: the table containing the columns
     """
     def __init__(self, table):
         self.table = table
+        # ``self._columns`` attribute stores the bound columns (columns that
+        # have a real name, )
+        self._columns = SortedDict()
+
+    def _spawn_columns(self):
+        """
+        (re)build the "_bound_columns" cache of :class:`.BoundColumn` objects
+        (note that :attr:`.base_columns` might have changed since last time);
+        creating :class:`.BoundColumn` instances can be costly, so we reuse
+        existing ones.
+        """
+        columns = SortedDict()
+        for name, column in self.table.base_columns.items():
+            if name in self._columns:
+                columns[name] = self._columns[name]
+            else:
+                columns[name] = BoundColumn(self.table, column, name)
+        self._columns = columns
 
     def iternames(self):
         return (name for name, column in self.iteritems())
@@ -589,24 +500,12 @@ class BoundColumns(object):
         """
         Return an iterator of ``(name, column)`` pairs (where ``column`` is a
         :class:`.BoundColumn` object).
-
-        This method is the mechanism for retrieving columns that takes into
-        consideration all of the ordering and filtering modifiers that a table
-        supports (e.g. ``exclude`` and ``sequence``).
         """
-        # First we build a sorted dict of all the columns that we need.
-        columns = SortedDict()
-        for name, column in self.table.base_columns.iteritems():
-            columns[name] = BoundColumn(self.table, column, name)
-
-        # A list of column names in the correct sequence that they should be
-        # rendered in the table.
-        sequence = (self.table.sequence or Sequence(('...', )))
-        sequence.expand(self.table.base_columns.keys())
-
-        for name in sequence:
-            if name not in self.table.exclude:
-                yield (name, columns[name])
+        self._spawn_columns()
+        if self.table.sequence:
+            return ((x, self._columns[x]) for x in self.table.sequence)
+        else:
+            return self._columns.iteritems()
 
     def items(self):
         return list(self.iteritems())
@@ -621,7 +520,7 @@ class BoundColumns(object):
         conjunction with e.g. ``{{ forloop.last }}`` (the last column might not
         be the actual last that is rendered).
         """
-        return ifilter(lambda x: x.sortable, self.iterall())
+        return ifilter(lambda x: x.sortable, self.all())
 
     def sortable(self):
         return list(self.itersortable())
@@ -633,14 +532,14 @@ class BoundColumns(object):
 
         This is geared towards table rendering.
         """
-        return ifilter(lambda x: x.visible, self.iterall())
+        return ifilter(lambda x: x.visible, self.all())
 
     def visible(self):
         return list(self.itervisible())
 
     def __iter__(self):
         """
-        Convenience API, alias of :meth:`.itervisible`.
+        Convenience API with identical functionality to :meth:`visible`.
         """
         return self.itervisible()
 
@@ -662,6 +561,7 @@ class BoundColumns(object):
         Return how many :class:`BoundColumn` objects are contained (and
         visible).
         """
+        self._spawn_columns()
         return len(self.visible())
 
     def __getitem__(self, index):
@@ -675,17 +575,11 @@ class BoundColumns(object):
             columns['speed']  # returns a bound column with name 'speed'
             columns[0]        # returns the first column
         """
+        self._spawn_columns()
         if isinstance(index, int):
-            try:
-                return next(islice(self.iterall(), index, index+1))
-            except StopIteration:
-                raise IndexError
+            return self._columns.value_for_index(index)
         elif isinstance(index, basestring):
-            for column in self.iterall():
-                if column.name == index:
-                    return column
-            raise KeyError("Column with name '%s' does not exist; "
-                           "choices are: %s" % (index, self.names()))
+            return self._columns[index]
         else:
-            raise TypeError(u'row indices must be integers or str, not %s'
-                            % index.__class__.__name__)
+            raise TypeError(u'row indices must be integers or str, not %s' %
+                            index.__class__.__name__)
